/********************************************************************
* ctlpanel.c                                                        *
*                                                                   *
* Description:                                                      *
*   Uses the curses library to paint a control panel, update        *
*   control displays, and accept keyboard control input to the      *
*   various fields.                                                 *
*                                                                   *
* Modification Record:                                              *
*                                                                   *
********************************************************************/
#include <curses.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "global.h"
#include "ctlpanel.h"

/************************** Local Symbols **************************/
/* Index into "win" array of pointers to the control panel windows */
enum WIN_INDEX {
    WIN_SW1,             /* Window for Switch 1 */
    WIN_SW2,             /* Window for Switch 2 */
    WIN_LED1,            /* Window for Single LED 1 */
    WIN_LED2,            /* Window for Single LED 2 */
    WIN_ROTENC,          /* Window for Rotary Encoder */
    WIN_7SEG,            /* Window for Seven-Segment Display */
    NUMWINS              /* Number of windows on control panel */
};

/* Index into "tip" array of strings */
enum TIP_INDEX {
    TIP_SCAN,            /* Tip for scanning the array */
    TIP_NUMERIC,         /* Tip for entering decimal number */
    TIP_NUMERIC_HEX,     /* Tip for entering hexadecimal number */
    TIP_NUMERIC_OCTAL,   /* Tip for entering octal number */
    TIP_ENTER,           /* Tip for just pressing <Enter> */
    TIP_CLEAR            /* Clear the field */
};

/* Index into "alert" array of strings */
enum ALERT_INDEX {
    ALERT_UPPER,         /* Alert that upper limit too small */
    ALERT_LOWER,         /* Alert that lower limit too big */
    ALERT_CLEAR          /* Clear the field */
};

/* Phase of operation of alert window */
enum ALERT_PHASE {
    ALERT_PHASE_NONE,    /* No alert on window */
    ALERT_PHASE_PRESENT, /* Alert on window */
    ALERT_PHASE_CLEAR    /* Alert ready to be cleared */
};

#define ROW1_STARTY  1   /* Line for start of first row of windows.
                          * Just makes room for screen header */
#define ROW2_STARTY  7   /* Next row.  Make room for max content
                          * lines for Row 1 windows.  See "content" below. */
#define ROW3_STARTY  12  /* Next row.  Make room for max content
                          * lines for Row 2 windows.  See "content" below. */
#define ROW4_STARTY  22  /* Comes after Row 3 windows; make room for max
                          * content lines for Row 2 windows.  See "content" 
                          * below.  No actual row of windows starts here,
                          * but this is where context-dependent help will
                          * be displayed. */

/* Color pair numbers, to be initialized with init_pair()
 * and used with COLOR_PAIR() */
#define PAIR_WHITE   1
#define PAIR_CYAN    2
#define PAIR_YELLOW  3
#define PAIR_GREEN   4

/* Window content strings */
/* stdscr header */
#define STDSCR_HEAD  \
"P A R P O R T   C O N T R O L   P A N E L"
#define STDSCR_FOOT  \
"<Up/Down> or <Tab> to change selection, <Q> to quit"

/* X-coordinate of display parameters.  Same for all windows.
 * Has to be larger than the longest content line in "content". */
#define PARAMX 17

/************************* Local Constants *************************/
/* Not all controls are "controllable"; some are merely displays.
 * "tab_order" is an array of controllable controls in the order
 * that the tab key will traverse them. */
enum PARPORT_CONTROL tab_order[] = {
    WIN_SW1_MODE,        /* Mode for Switch 1 */
    WIN_SW2_MODE,        /* Mode for Switch 2 */
    WIN_LED1_MODE,       /* Mode for Single LED 1 */
    WIN_LED2_MODE,       /* Mode for Single LED 2 */
    WIN_ROTENC_RESET,    /* Control to clear Rotary Encoder count */
    WIN_ROTENC_ULIMIT,   /* Rotary Encoder count upper limit */
    WIN_ROTENC_LLIMIT,   /* Rotary Encoder count lower limit */
    WIN_ROTENC_BASE,     /* Counting Base for Rotary Encoder */
    WIN_ROTENC_MODE,     /* Mode for Rotary Encoder */
    WIN_7SEG_BASE,       /* Counting Base for Seven-Segment Display */
    WIN_7SEG_MODE,       /* Mode for Seven-Segment Display */
    MAX_PARPORT_CONTROL  /* Marks the end of tab_order */
};

/* Arrays of text strings corresponding to enums in "global.h" */
const char *text_sw1_mode[] = {      /* Switch 1 Mode */
    "Normal",
    "Indicator 1",
    "7-Seg Mode",
    "Disabled"
};

const char *text_sw2_mode[] = {      /* Switch 2 Mode */
    "Normal",
    "Indicator 2",
    "Blank Displays",
    "Quit Program",
    "Disabled"
};

const char *text_led1_mode[] = {     /* Indicator 1 Mode */
    "Normal",
    "Blink",
    "Soft Blink",
    "Always On",
    "Always Off"
};

const char *text_led2_mode[] = {     /* Indicator 2 Mode */
    "Normal",
    "Blink",
    "Soft Blink",
    "Always On",
    "Always Off"
};

const char *text_rotenc_mode[] = {   /* Rotary Encoder Mode */
    "Normal",
    "Limited",
    "Disabled"
};

const char *text_7seg_mode[] = {     /* Seven-Segment Display Mode */
    "Timer",
    "Rotary Encoder",
    "Lamp Test",
    "Blank"
};

const char *base_name[] = {          /* Number base */
    "Decimal",
    "Hexadecimal",
    "Octal"
};

const char *base_format[] = {        /* Format string for selected base */
    "%02d",               /* DECIMAL, Base-10 */
    "%02X",               /* HEXADECIMAL, Base-16 */
    "%02o"                /* OCTAL, Base-8 */
};

const char *tip[] = {                /* "tip" array of strings */
    "<Left/Right> to view options, <Enter> to select", /* TIP_SCAN */
    "Digits (0-9), <Enter> to select",         /* TIP_NUMERIC */
    "Digits (0-9, A-F), <Enter> to select",    /* TIP_NUMERIC_HEX */
    "Digits (0-7), <Enter> to select",         /* TIP_NUMERIC_OCTAL */
    "Press <Enter> to take action",            /* TIP_ENTER */
    ""                                         /* TIP_CLEAR */
};

const char *alert[] = {              /* "alert" array of strings */
    "Upper limit cannot be less than lower limit",   /* ALERT_UPPER */
    "Lower limit cannot be more than upper limit",   /* ALERT_LOWER */
    ""                                               /* ALERT_CLEAR */
};

/* Database of the contents of all the windows.
 * First part of array corresponds with PARPORT_CONTROL indices
 * (see "global.h"). */
struct WIN_CONTENT {
    enum WIN_INDEX i_win;        /* Window to which this content goes */
    int  y;                      /* Line for content */
    int  x;                      /* Column for content */
    int  color;                  /* Color to paint content */
    char *str;                   /* Content string itself */
} content[] = {
    /* WIN_SW1_STATE */
    { WIN_SW1,    2,  1,  PAIR_GREEN, "Toggle State :" },
    /* WIN_SW1_MODE */
    { WIN_SW1,    3,  1,  PAIR_GREEN, "Mode :" },
    /* WIN_SW2_STATE */
    { WIN_SW2,    2,  1,  PAIR_GREEN, "Toggle State :" },
    /* WIN_SW2_MODE */
    { WIN_SW2,    3,  1,  PAIR_GREEN, "Mode :" },
    /* Row 1 max lines: 6 */
    /* WIN_LED1_MODE */
    { WIN_LED1,   2,  1,  PAIR_GREEN, "Mode :" },
    /* WIN_LED2_MODE */
    { WIN_LED2,   2,  1,  PAIR_GREEN, "Mode :" },
    /* Row 2 max lines: 5 */
    /* WIN_ROTENC_COUNT */
    { WIN_ROTENC, 2,  1,  PAIR_GREEN, "Count :" },
    /* WIN_ROTENC_RESET */
    { WIN_ROTENC, 3,  1,  PAIR_GREEN, "Reset Count :" },
    /* WIN_ROTENC_ULIMIT */
    { WIN_ROTENC, 4,  1,  PAIR_GREEN, "Upper Limit :" },
    /* WIN_ROTENC_LLIMIT */
    { WIN_ROTENC, 5,  1,  PAIR_GREEN, "Lower Limit :" },
    /* WIN_ROTENC_BASE */
    { WIN_ROTENC, 6,  1,  PAIR_GREEN, "Count Base :" },
    /* WIN_ROTENC_MODE */
    { WIN_ROTENC, 7,  1,  PAIR_GREEN, "Mode :" },
    /* WIN_7SEG_VALUE */
    { WIN_7SEG,   2,  1,  PAIR_GREEN, "Display :" },
    /* WIN_7SEG_BASE */
    { WIN_7SEG,   3,  1,  PAIR_GREEN, "Display Base :" },
    /* WIN_7SEG_MODE */
    { WIN_7SEG,   4,  1,  PAIR_GREEN, "Mode :" },
    /* Row 3 max lines: 10 */
    /* Content not associated with controls */
    { WIN_SW1,    0,  4,  PAIR_CYAN, "Switch 1" },
    { WIN_SW2,    0,  4,  PAIR_CYAN, "Switch 2" },
    { WIN_LED1,   0,  4,  PAIR_CYAN, "Indicator 1" },
    { WIN_LED2,   0,  4,  PAIR_CYAN, "Indicator 2" },
    { WIN_ROTENC, 0,  4,  PAIR_CYAN, "Rotary Encoder" },
    { WIN_7SEG,   0,  4,  PAIR_CYAN, "Seven-Segment Display" },
    { NUMWINS,    0,  0,  0,  "" }    /* End of array marker */
};

/* Database of the descriptions of all the controls.  Note that the
 * window and location (line) of the control display will be in the
 * "content" array.  These items are also in order of the PARPORT_CONTROL
 * indices in "global.h". */
struct CONTROL_DESC {
    int *pval;                   /* Points to parameter variable */
    enum TIP_INDEX tip;          /* Tip associated with control.  Of
                                  * necessity, also what kind of control. */
    void *pmisc;                 /* For TIP_SCAN control types, points to 
                                  * option string array.  For TIP_NUMERIC
                                  * control types, points to base control. */
    int maxval;                  /* Upper limit on the variable.  Really
                                    only useful for TIP_SCAN controls. */
} control[] = {
    /* WIN_SW1_STATE */
    { (int *)&toggle_sw1,   TIP_CLEAR,   NULL,             0 },
    /* WIN_SW1_MODE */
    { (int *)&mode_sw1,     TIP_SCAN,    text_sw1_mode,    MAX_SW1_MODE },
    /* WIN_SW2_STATE */
    { (int *)&toggle_sw2,   TIP_CLEAR,   NULL,             0 },
    /* WIN_SW2_MODE */
    { (int *)&mode_sw2,     TIP_SCAN,    text_sw2_mode,    MAX_SW2_MODE },
    /* WIN_LED1_MODE */
    { (int *)&mode_led1,    TIP_SCAN,    text_led1_mode,   MAX_LED1_MODE },
    /* WIN_LED2_MODE */
    { (int *)&mode_led2,    TIP_SCAN,    text_led2_mode,   MAX_LED2_MODE },
    /* WIN_ROTENC_COUNT */
    { (int *)&count_rotenc, TIP_NUMERIC, &base_rotenc,     0 },
    /* WIN_ROTENC_RESET */
    { (int *)NULL,          TIP_ENTER,   NULL,             0 },
    /* WIN_ROTENC_ULIMIT */
    { (int *)&ulim_rotenc,  TIP_NUMERIC, &base_rotenc,     0 },
    /* WIN_ROTENC_LLIMIT */
    { (int *)&llim_rotenc,  TIP_NUMERIC, &base_rotenc,     0 },
    /* WIN_ROTENC_BASE */
    { (int *)&base_rotenc,  TIP_SCAN,    base_name,        MAX_COUNT_BASE },
    /* WIN_ROTENC_MODE */
    { (int *)&mode_rotenc,  TIP_SCAN,    text_rotenc_mode, MAX_ROTENC_MODE },
    /* WIN_7SEG_VALUE */
    { (int *)&display_7seg, TIP_NUMERIC, &base_7seg,       0 },
    /* WIN_7SEG_BASE */
    { (int *)&base_7seg,    TIP_SCAN,    base_name,        MAX_COUNT_BASE },
    /* WIN_7SEG_MODE */
    { (int *)&mode_7seg,    TIP_SCAN,    text_7seg_mode,   MAX_SVNSEG_MODE }
};

/******************* Local Persistant Variables ********************/
/* Curses Windows.  There are twice as many as required, since each
 * window will have a frame.  The frame window has a box drawn around
 * it and nothing else.  The controls window fits inside.  This way,
 * whatever happens to the controls window will not alter the box
 * around it. */
static WINDOW *win[NUMWINS];   /* Array of pointers to panel windows */
static WINDOW *frame[NUMWINS]; /* Array of pointers to panel frames */
static WINDOW *wintip;      /* Pointer for "Tip" window */
static WINDOW *winalert;    /* Pointer for "Alert" window */
static int active_control;  /* Index of active control in "tab_order[]" */
static int tentative;       /* Value of active control after change activity */
static int changed;         /* TRUE if content of "tentative" is valid */
static int *pparam;         /* Pointer to active control parameter */
static enum ALERT_PHASE alertphase;      /* Phase of alert window */

/************************ Local Prototypes *************************/
void move_to_next(int dir);
void scan_option(int adv);
void numeric_field(int c);
enum PARPORT_CONTROL accept_action(void);

/********************************************************************
* init_ctlpanel()                                                   *
*                                                                   *
* Description:                                                      *
*   Initialize curses system, create window set, and paint default  *
*   window contents.                                                *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
void init_ctlpanel(void)
{
    int i;
    int leny, lenx, starty, startx;
    enum PARPORT_CONTROL ctl;

    /* Ultimately, the first control will be the starting active one. */
    active_control = 0;
    tentative = 0;
    changed = 0;
    alertphase = ALERT_PHASE_NONE;

    /* Initialize curses */
    initscr();
    cbreak();
    noecho();
    curs_set(0);        /* Don't display text cursor */

    /* Set up colors for different window features */
    start_color();
    init_pair(PAIR_WHITE, COLOR_WHITE, COLOR_BLACK);
    init_pair(PAIR_CYAN, COLOR_CYAN, COLOR_BLACK);
    init_pair(PAIR_YELLOW, COLOR_YELLOW, COLOR_BLACK);
    init_pair(PAIR_GREEN, COLOR_GREEN, COLOR_BLACK);

    /* Put legend on base window (stdscr) and refresh before
     * creating and painting sub-windows */
    mvaddstr(0, (COLS - strlen(STDSCR_HEAD))/2, STDSCR_HEAD);
    mvaddstr(LINES-1, 0, STDSCR_FOOT);
    refresh();

    /* Create windows.  This is where the dimensions are specified. */
    for(i = WIN_SW1; i < NUMWINS; i++) {
        switch(i) {
            case WIN_SW1 :
                leny = ROW2_STARTY - ROW1_STARTY;
                lenx = COLS/2-1;
                starty = ROW1_STARTY;
                startx = 0;
                break;
            case WIN_SW2 :
                leny = ROW2_STARTY - ROW1_STARTY;
                lenx = COLS/2-1;
                starty = ROW1_STARTY;
                startx = COLS/2;
                break;
            case WIN_LED1 :
                leny = ROW3_STARTY - ROW2_STARTY;
                lenx = COLS/2-1;
                starty = ROW2_STARTY;
                startx = 0;
                break;
            case WIN_LED2 :
                leny = ROW3_STARTY - ROW2_STARTY;
                lenx = COLS/2-1;
                starty = ROW2_STARTY;
                startx = COLS/2;
                break;
            case WIN_ROTENC :
                leny = ROW4_STARTY - ROW3_STARTY;
                lenx = COLS/2-1;
                starty = ROW3_STARTY;
                startx = 0;
                break;
            case WIN_7SEG :
                leny = ROW4_STARTY - ROW3_STARTY;
                lenx = COLS/2-1;
                starty = ROW3_STARTY;
                startx = COLS/2;
                break;
        }
        /* Create the frame window, draw the box, and put it on the screen */
        frame[i] = newwin(leny, lenx, starty, startx);
        wattron(frame[i], COLOR_PAIR(PAIR_YELLOW));
        box(frame[i], 0 , 0);
        wrefresh(frame[i]);
        /* Create the nested controls window and enable for keypad input */
        win[i] = newwin(leny-2, lenx-2, starty+1, startx+1);
        keypad(win[i], TRUE);
        nodelay(win[i], TRUE);
    }

    /* Load windows with initial text content. */
    for(i = 0; content[i].i_win != NUMWINS; i++) {
        wattron(win[content[i].i_win], COLOR_PAIR(content[i].color));
        mvwaddstr(win[content[i].i_win],
                content[i].y, content[i].x, content[i].str);
    }

    /* Now, put up all the control parameters, and highlight the
     * starting-off active one. 
     * Note that "update_ctlpanel()" includes a wrefresh() for
     * each window, so nothing more needs to be done. */
    for(ctl = WIN_SW1_STATE; ctl < MAX_PARPORT_CONTROL; ctl++) {
        update_ctlpanel(ctl);
    }

    /* Create windows for tip and alert.  Put up tip for active control. */
    wintip   = newwin(1, 0, ROW4_STARTY,   0);
    winalert = newwin(1, 0, ROW4_STARTY+1, 0);
    mvwaddstr(wintip, 0, 0, tip[control[tab_order[active_control]].tip]);
    wrefresh(wintip);
}

/********************************************************************
* shutdown_ctlpanel()                                               *
*                                                                   *
* Description:                                                      *
*   Destroy windows, free allocation, and stop curses.              *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
void shutdown_ctlpanel(void)
{
    /* Free up our curses allocated storage */
    endwin();
}

/********************************************************************
* process_ctlpanel()                                                *
*                                                                   *
* Description:                                                      *
*   Accept keypresses from console, and take appropriate action.    *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : Return values as defined in global.h                    *
********************************************************************/
enum PARPORT_CONTROL process_ctlpanel(void)
{
    int c;
    enum PARPORT_CONTROL retval;

    retval = CTLPANEL_NOACTION;

    /* Get a keypress from the console.  Have to use "wgetch" because
     * keypad was enabled only for windows.  Make sure all lower-case
     * letters are converted to upper-case. Have to do this with an 
     * "if" because DJGPP "toupper" messes with keypad codes. */
    c = wgetch(win[0]);
    if(c >= 'a' && c <= 'z') {
        c = toupper(c);
    }

    /* Process keypress */
    switch(c) {
        case ERR :        /* No keypress.  Just quit */
            break;
        case 0x0D :       /* Enter key.  Accept input (if changed) */
        case 0x0A :
            retval = accept_action();
            /* If a control entry error occurs, don't move */
            if(retval == MAX_PARPORT_CONTROL) {
                move_to_next(0);
                retval = CTLPANEL_NOACTION;
            }
            else {
                move_to_next(1);
            }
            break;
        case KEY_DOWN :   /* Down arrow.  Move to next controllable field */
        case 9 :          /* <Tab> does the same thing */
            move_to_next(1);
            break;
        case KEY_UP:      /* Up arrow.  Move to previous controllable field */
            move_to_next(-1);
            break;
        case KEY_RIGHT :  /* Right arrow.  Scan options "forward" */
            scan_option(1);
            break;
        case KEY_LEFT :   /* Left arrow.  Scan options "backward" */
            scan_option(-1);
            break;
        case 'Q' :        /* Q, for Quit */
            retval = CTLPANEL_TERMINATE;
            break;
        default :
            if(c >= '0' && c <= '9') {
                c = c - '0';
                numeric_field(c);
            }
            else if (c >= 'A' && c <= 'F') {
                c = c - 'A' + 0x0A;
                numeric_field(c);
            }
            break;
    }

    return retval;
}

/********************************************************************
* update_ctlpanel()                                                 *
*                                                                   *
* Description:                                                      *
*   Update designated control from corresponding global variable.   *
*                                                                   *
* Inputs  : enum PARPORT_CONTROL ctl, control to update.            *
* Outputs : None.                                                   *
********************************************************************/
void update_ctlpanel(enum PARPORT_CONTROL ctl)
{
    int *pval;              /* Points to the parameter value. */
    const char **ptext;     /* Points to array of TIP_SCAN strings */
    enum COUNT_BASE base;   /* Base of numeric parameters */
    int select;             /* TRUE if control being updated is the
                             * active control (and therefore should be
                             * highlighted); else FALSE */

    select = (ctl == tab_order[active_control]);
    pval = control[ctl].pval;

    /* If change activity is underway, use the tentative value for
     * display instead of the original. */
    if(select && changed) {
        pval = &tentative;
    }

    /* Turn on highlighting if requested */
    if(select) {
        wattron(win[content[ctl].i_win], A_REVERSE);
    }
    wattron(win[content[ctl].i_win], COLOR_PAIR(PAIR_WHITE));

    /* Print to screen with format depending on control type */
    switch(control[ctl].tip) {
        case TIP_SCAN :
            ptext = (const char **)control[ctl].pmisc;
            mvwprintw(win[content[ctl].i_win],
                    content[ctl].y, PARAMX,
                    ptext[*pval]);
            break;
        case TIP_NUMERIC :
            base = *((enum COUNT_BASE *)control[ctl].pmisc);
            mvwprintw(win[content[ctl].i_win],
                    content[ctl].y, PARAMX,
                    base_format[base], *pval);
            break;
        case TIP_ENTER :
            mvwprintw(win[content[ctl].i_win],
                    content[ctl].y, PARAMX,
                    "<Enter>");
            break;
        case TIP_CLEAR :      /* Special Cases */
            if(ctl == WIN_SW1_STATE || ctl == WIN_SW2_STATE) {
                mvwprintw(win[content[ctl].i_win],
                        content[ctl].y, PARAMX,
                        "%d", *pval);
            }
            break;
        default :       /* Never get here.  Makes the compiler happy */
            break;
    }
    /* Clean anything residual from this line */
    wclrtoeol(win[content[ctl].i_win]);
    /* Now turn it off again */
    if(select) {
        wattroff(win[content[ctl].i_win], A_REVERSE);
    }

    /* Update window on screen just before leaving */
    wrefresh(win[content[ctl].i_win]);
}

/********************************************************************
* move_to_next()                                                    *
*                                                                   *
* Description:                                                      *
*   Move the highlighted selection to the next control in           *
*   "tab_order".                                                    *
*                                                                   *
* Inputs  : int dir, direction to move (positive down).  If zero,   *
*             don't move anywhere, just refresh the window.         *
* Outputs : None.                                                   *
********************************************************************/
void move_to_next(int dir)
{
    int old_control;

    changed = FALSE;    /* Abandon any uncommitted changes */
    old_control = active_control;
    if(dir > 0) {
        active_control++;
        if(tab_order[active_control] == MAX_PARPORT_CONTROL) {
            active_control = 0;
        }
    }
    else if(dir < 0) {
        if(active_control == 0) {
            while(tab_order[active_control] != MAX_PARPORT_CONTROL) {
                active_control++;
            }
        }
        active_control--;
    }

    /* If some function has put an alert on the screen in response
     * to a keypress, leave it up after the <Enter>, but remove
     * it after the next movement */
    switch(alertphase) {
        case ALERT_PHASE_NONE :    /* No message; do nothing */
            break;
        case ALERT_PHASE_PRESENT : /* First time; just advance phase */
            alertphase = ALERT_PHASE_CLEAR;
            break;
        case ALERT_PHASE_CLEAR :   /* Second time; remove message */
            mvwaddstr(winalert, 0, 0, alert[ALERT_CLEAR]);
            wclrtoeol(winalert);
            wrefresh(winalert);
            alertphase = ALERT_PHASE_NONE;
            break;
    }

    /* Turn off highlighting for old control */
    update_ctlpanel(tab_order[old_control]);
    /* Turn on highlighting for new control */
    update_ctlpanel(tab_order[active_control]);
    /* Display tip for new control */
    mvwaddstr(wintip, 0, 0, tip[control[tab_order[active_control]].tip]);
    wclrtoeol(wintip);
    wrefresh(wintip);
}

/********************************************************************
* scan_option()                                                     *
*                                                                   *
* Description:                                                      *
*   Traverse the options of an enumerated control, and update the   *
*   window accordingly.                                             *
*                                                                   *
* Inputs  : int adv, increment to increase or decrement option index*
* Outputs : None.                                                   *
********************************************************************/
void scan_option(int adv)
{
    int *pval;            /* Point to variable to be adjusted */
    int max;              /* Contains max in range */

    /* If the active control is not a numeric parameter, just quit */
    if(control[tab_order[active_control]].tip != TIP_SCAN) {
        return;
    }

    /* First, limit adv to a single increment or decrement */
    if(adv < 0) {
        adv = -1;
    }
    else {
        adv = 1;
    }

    /* Perform increment or decrement to appropriate variable, and
     * roll at the ends of the range. */
    pval = control[tab_order[active_control]].pval;
    max = control[tab_order[active_control]].maxval;

    /* Capture the original parameter value and the fact that it
     * is being changed */
    if(!changed) {
        tentative = *pval;
        pparam = pval;
        changed = TRUE;
    }
    tentative += adv;

    /* Perform roll-over at ends of range */
    if(tentative == max) {
        tentative = 0;
    }
    if(tentative < 0) {
        tentative = max-1;
    }
    /* Now do the screen update */
    update_ctlpanel(tab_order[active_control]);
}

/********************************************************************
* numeric_field()                                                   *
*                                                                   *
* Description:                                                      *
*   If the active control is a numeric field, and the received      *
*   number is within the base for the field, shift the              *
*   number in as the new least-significant digit.                   *
*                                                                   *
* Inputs  : int c, prospective least-significant digit              *
* Outputs : None.                                                   *
********************************************************************/
void numeric_field(int c)
{
    int *pval;            /* Point to variable to be adjusted */
    enum COUNT_BASE base; /* Numeric base for selected parameter */
    enum COUNT_BASE *pbase; /* Pointer to source for base */

    /* If the active control is not a numeric parameter, just quit */
    if(control[tab_order[active_control]].tip != TIP_NUMERIC) {
        return;
    }

    /* Get the pointer to the parameter, and the number base */
    pval = control[tab_order[active_control]].pval;
    pbase = (enum COUNT_BASE *)control[tab_order[active_control]].pmisc;
    base = *pbase;

    /* Okay, a valid numeric field is active.  Check the prospective
     * least-significant digit for the base, and if not, just quit */
    if(((base == DECIMAL) && !(c >= 0 && c <= 9)) ||
            ((base == HEXADECIMAL) &&
                !((c >= 0 && c <= 9) || (c >= 0x0A && c <= 0x0F))) ||
            ((base == OCTAL) && !(c >= 0 && c <= 7))) {
        return;
    }

    /* If the first input to this field, grab the current parameter
     * value and indicate a change is underway */
    if(!changed) {
        tentative = *pval;
        pparam = pval;
        changed = TRUE;
    }
    tentative = tentative * base_digit[base] + c;
    tentative = tentative % base_max[base];
    /* Now do the screen update */
    update_ctlpanel(tab_order[active_control]);
}

/********************************************************************
* accept_action()                                                   *
*                                                                   *
* Description:                                                      *
*   Upon <Enter>, if a change is outstanding, check for any error   *
*   conditions.  If so, discard the change and report the error;    *
*   otherwise, accept the update.                                   *
*                                                                   *
* Inputs  : None.                                                   *
* Outputs : PARPORT_CONTROL identifier of changed control.  If an   *
*             error occurred, return MAX_PARPORT_CONTROL            *
********************************************************************/
enum PARPORT_CONTROL accept_action(void)
{
    /* Controls that are TIP_ENTER are always accepted */
    if(control[tab_order[active_control]].tip == TIP_ENTER) {
        return tab_order[active_control];
    }

    /* If no change, do nothing more */
    if(!changed) {
        return CTLPANEL_NOACTION;
    }

    /* Perform special processing associated with this control */
    switch(tab_order[active_control]) {
        case WIN_ROTENC_ULIMIT :
            /* Proposed upper limit cannot be less than lower limit */
            if(tentative < llim_rotenc) {
                mvwaddstr(winalert, 0, 0, alert[ALERT_UPPER]);
                wclrtoeol(winalert);
                wrefresh(winalert);
                alertphase = ALERT_PHASE_PRESENT;
                return MAX_PARPORT_CONTROL;
            }
            break;
        case WIN_ROTENC_LLIMIT :
            /* Proposed lower limit cannot be more than upper limit */
            if(tentative > ulim_rotenc) {
                mvwaddstr(winalert, 0, 0, alert[ALERT_LOWER]);
                wclrtoeol(winalert);
                wrefresh(winalert);
                alertphase = ALERT_PHASE_PRESENT;
                return MAX_PARPORT_CONTROL;
            }
            break;
        default :       /* No other controls have special handling */
            break;
    }
    *pparam = tentative;
    /* return the control ID for the changed control */
    return tab_order[active_control];
}

