/********************************************************************
* lpt_ctrl.c                                                        *
*                                                                   *
* Description:                                                      *
*   Controls a module connected to the LPT1 parallel port with the  *
*   following connections:                                          *
*     D0 - D6 output    : Segment anodes of two seven-segment LEDs  *
*     D7 output         : Anodes of two LEDs                        *
*     STROBE output     : Cathode of rightmost seven-segment LED    *
*     LINEFEED output   : Cathode of leftmost seven-segment LED     *
*     SELECT_OUT output : Cathode of rightmost single LED           *
*     INIT output       : Cathode of leftmost single LED            *
*     ERROR input       : Sense side of rightmost pushbutton switch *
*     SELECT_IN input   : Sense side of leftmost pushbutton switch  *
*     PAPER_OUT input   : Channel A of rotary encoder               *
*     ACK input         : Channel B of rotary encoder               *
*                                                                   *
* Modification Record:                                              *
*                                                                   *
********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#ifdef INTERRUPT
    #include <dpmi.h>
    #include <go32.h>
#endif   /* INTERRUPT */
#ifndef MINGW
    #include <pc.h>
#endif
#include <dos.h>
#include "global.h"
#include "lpt_ctrl.h"

/********************************************************************
* Register Values
*   PORT - Offset 0, 8-bit output to seven-segment and LED drive
*   STATUS - Offset 1, 8-bit input
*   CONTROL - Offset 2, 8-bit output (can be read as inputs)
*   Status and Control pins normally have pull-up resistors
*
* Status Inputs:
*   Busy (status 7) reads inverted
*   Ack (status 6) reads true
*   Paper Out (status 5) reads true
*   Select In (status 4) reads true
*   Error (status 3) reads true
* Control Outputs:
*   Select Out (control 3) outputs inverted
*   Initialize (control 2) outputs true
*   Linefeed (control 1) outputs inverted
*   Strobe (control 0) outputs inverted
* Control Inputs:
*   Select Out (control 3) reads inverted
*   Initialize (control 2) reads true
*   Linefeed (control 1) reads inverted
*   Strobe (control 0) reads inverted
********************************************************************/

/************************** Local Symbols **************************/
#define ADRS    0x378     /* I/O address for parallel port */
#define PORT    (ADRS)    /* I/O address for PORT register */
#define STATUS  (ADRS+1)  /* I/O address for STATUS register */
#define CONTROL (ADRS+2)  /* I/O address for CONTROL register */

/* Definitions for Seven-Segment LED displays */
#define SEG_A  0x01    /*   -- SEG_F --      */
#define SEG_B  0x02    /*  |           |     */
#define SEG_C  0x04    /* SEG_A      SEG_E   */
#define SEG_D  0x08    /*  |           |     */
#define SEG_E  0x10    /*   -- SEG_G --      */
#define SEG_F  0x20    /*  |           |     */
#define SEG_G  0x40    /* SEG_B      SEG_D   */
                       /*  |           |     */
                       /*   -- SEG_C --      */

#define CHAR_1  (SEG_D+SEG_E)
#define CHAR_2  (SEG_C+SEG_B+SEG_G+SEG_E+SEG_F)
#define CHAR_3  (SEG_C+SEG_D+SEG_G+SEG_E+SEG_F)
#define CHAR_4  (SEG_D+SEG_E+SEG_G+SEG_A)
#define CHAR_5  (SEG_C+SEG_D+SEG_G+SEG_A+SEG_F)
#define CHAR_6  (SEG_B+SEG_C+SEG_D+SEG_G+SEG_A+SEG_F)
#define CHAR_7  (SEG_D+SEG_E+SEG_F)
#define CHAR_8  (SEG_A+SEG_B+SEG_C+SEG_D+SEG_E+SEG_F+SEG_G)
#define CHAR_9  (SEG_C+SEG_D+SEG_E+SEG_F+SEG_A+SEG_G)
#define CHAR_0  (SEG_A+SEG_B+SEG_C+SEG_D+SEG_E+SEG_F)
#define CHAR_A  (SEG_A+SEG_B+SEG_D+SEG_E+SEG_F+SEG_G)
#define CHAR_B  (SEG_A+SEG_B+SEG_C+SEG_D+SEG_G)
#define CHAR_C  (SEG_B+SEG_C+SEG_G)
#define CHAR_D  (SEG_B+SEG_C+SEG_D+SEG_E+SEG_G)
#define CHAR_E  (SEG_A+SEG_B+SEG_C+SEG_F+SEG_G)
#define CHAR_F  (SEG_A+SEG_B+SEG_F+SEG_G)

/* LED high-side drive in PORT register (works in concert with
 * low-side CONTROL register drivers */
#define LED_DRIVE 0x80

/* Bit fields for STATUS port */
#define RSWITCH  0x08      /* Rightmost switch (2) */
#define LSWITCH  0x10      /* Leftmost switch (1) */
#define SWITCHES 0x18      /* Isolate switch bits */
#define ROTENC_A 0x20      /* Channel A of Rotary Encoder */
#define ROTENC_B 0x40      /* Channel C of Rotary Encoder */
#define ROTENCS  0x60      /* Isolate Rotary Encoder bits */

/* Bit fields for CONTROL port */
#define RSEVSEG  0x01      /* Driver for rightmost 7-segment */
#define LSEVSEG  0x02      /* Driver for leftmost 7-segment */
#define SEVSEGS  0x03      /* Isolate 7-seg driver bits */
#define RLEDIND  0x08      /* Driver for rightmost LED (2) */
#define LLEDIND  0x04      /* Driver for leftmost LED (1) */
#define LEDINDS  0x0C      /* Isolate LED bits */
#define INT_ENABLE  0x10   /* ACK Interrupt Enable */

#ifdef INTERRUPT
/* DJGPP Interrupt Service Routine symbols/macros */
#define IRQ 7               /* IRQ Here */
#define PIC1 0x20           /* Port addresses of Programmable Interrupt */
#define PIC2 0xA0           /*   Controllers 1 and 2 */

/* Macros by Shawn Hargreaves from the Allegro library for locking
 * functions and variables. */
#define LOCK_VARIABLE(x)    _go32_dpmi_lock_data((void *)&x,(long)sizeof(x));
#define LOCK_FUNCTION(x)    _go32_dpmi_lock_code(x,(long)sizeof(x));
#else    /* Non-Interrupt */
enum ROTARY_STATE {
    RSTATE_0,               /* Quiescent state */
    RSTATE_1L,              /* First State for Leftward turn */
    RSTATE_2L,              /* Second State for Leftward turn */
    RSTATE_3L,              /* Third State for Leftward turn */
    RSTATE_1R,              /* First State for Rightward turn */
    RSTATE_2R,              /* Second State for Rightward turn */
    RSTATE_3R               /* Third State for Rightward turn */
};
#endif   /* INTERRUPT */

/************************* Local Constants *************************/
/* Bit patterns for PORT for the seven-segment LED displays */
unsigned char cycle[16] = {
    CHAR_0, CHAR_1, CHAR_2, CHAR_3, CHAR_4,
    CHAR_5, CHAR_6, CHAR_7, CHAR_8, CHAR_9,
    CHAR_A, CHAR_B, CHAR_C, CHAR_D, CHAR_E, CHAR_F
};

/******************* Local Persistant Variables ********************/
static int shadow_control;      /* Shadow of CONTROL register */
static int shadow_port;         /* Shadow of PORT register */
static int sw1_debounce;        /* Debounce counter for Switch 1 */
static int sw2_debounce;        /* Debounce counter for Switch 2 */

#ifdef INTERRUPT
/* Variables associated with the Interrupt Service Routine.  The first
 * ones will get their memory assignments locked down. */
static int isrflag;             /* Interrupt Flag */
static int dirflag;             /* Direction Flag */
static int picaddr;             /* Programmable Interrupt Controller (PIC)
                                 * Base Address */
static int intno;               /* Interrupt Number */
static int picmask;             /* PIC's Mask */
static _go32_dpmi_seginfo OldISR; /* Old ISR context */
static _go32_dpmi_seginfo NewISR; /* New ISR context */
#else    /* Non-Interrupt */
enum ROTARY_STATE rotary_state;
#endif   /* INTERRUPT */

/************************ Local Prototypes *************************/
#ifdef INTERRUPT
void ISR_lpt(void);
#endif   /* INTERRUPT */

#ifdef MINGW
/* Windows doesn't allow port access.  Use io.dll functions.
 * Obtain io.dll and source files from http://geekhideout.com/iodll.shtml
 * and use Makefile.mingw.
 * Note that io.dll has to be moved to c:\windows\system32 to work.
 * Pull in source file that sets up io.dll functions for use,
 * and add some outportb and inportb wrappers. */
#include "io.cpp"
inline void outportb(int a, int d)
{
    PortOut(a, d);
}
inline int inportb(int a)
{
    return PortIn(a);
}
#endif

/********************************************************************
* init_parport()                                                    *
*                                                                   *
* Description:                                                      *
*   Initialize variables and hardware registers associated with the *
*   parallel port control CSC.                                      *
*   Note: A hardware Interrupt Service Routine is installed by this *
*   function.  To properly stop the ISR, "shutdown_parport()" must  *
*   be called before the program terminates.                        *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
void init_parport(void)
{
#ifdef MINGW
    LoadIODLL();
#endif
    /* All PORT bits off except for LED high-side */
    shadow_port = LED_DRIVE;
    /* All CONTROL bits off (note LED 1 driver is NOT inverted */
    shadow_control = LLEDIND;

    sw1_debounce = 0;
    sw2_debounce = 0;

#ifdef INTERRUPT
    /* Set up for interrupts */
    /* Calculate Interrupt Number, PIC Addr & Mask. */
    if(IRQ >= 2 && IRQ <= 7) {
        intno = IRQ + 0x08;
        picaddr = PIC1;
        picmask = 1;
        picmask = picmask << IRQ;
    }
    if(IRQ >= 8 && IRQ <= 15) {
        intno = IRQ + 0x68;
        picaddr = PIC2;
        picmask = 1;
        picmask = picmask << (IRQ-8);
    }
    if(IRQ < 2 || IRQ > 15) {
        printf("IRQ Out of Range\n");
        exit(0);
    }

    /* lock the functions and variables */
    LOCK_FUNCTION(ISR_lpt);
    LOCK_VARIABLE(isrflag);
    LOCK_VARIABLE(dirflag);
    LOCK_VARIABLE(picaddr);

    /* load the address of the old timer ISR into the OldISR structure */
    _go32_dpmi_get_protected_mode_interrupt_vector(intno, &OldISR);
    /* point NewISR to the proper selector:offset for handler
    * function */
    NewISR.pm_offset = (int)ISR_lpt;
    NewISR.pm_selector = _go32_my_cs();
    _go32_dpmi_allocate_iret_wrapper(&NewISR);
    _go32_dpmi_set_protected_mode_interrupt_vector(intno,&NewISR);
    /* Un-Mask PIC and enable Parallel Port IRQ */
    outportb(picaddr+1,inportb(picaddr+1) & (0xFF - picmask));
    shadow_control |= INT_ENABLE;
    isrflag = 0;
    dirflag = 0;
#else    /* Non-Interrupt */
    rotary_state = RSTATE_0;
#endif   /* INTERRUPT */

    outportb(PORT, shadow_port);
    outportb(CONTROL, shadow_control);
}

/********************************************************************
* shutdown_parport()                                                *
*                                                                   *
* Description:                                                      *
*   Uninstalls ISR for ACK interrupt, and puts parallel port        *
*   hardware registers in quiescent state.                          *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
void shutdown_parport(void)
{
    /* All outputs off (note LED 1 driver is NOT inverted */
    shadow_port = 0;
    shadow_control = (SEVSEGS + LEDINDS) & ~LLEDIND;
    outportb(PORT, shadow_port);
    outportb(CONTROL, shadow_control);
    
#ifdef INTERRUPT
    /* Disable Parallel Port IRQ */
    outportb(picaddr+1,inportb(picaddr+1) | picmask); /* Mask Pic */
    /* load the old timer ISR back without the new ISR chained on */
    _go32_dpmi_set_protected_mode_interrupt_vector(intno,&OldISR);
#endif   /* INTERRUPT */
}

#ifdef INTERRUPT
/********************************************************************
* ISR_lpt()                                                         *
*                                                                   *
* Description:                                                      *
*   Interrupt Service Routine.  Capture the fact that the interrupt *
*   occurred, and the direction.  Channel B is the interrupt        *
*   (positive edge), so Channel A indicates the direction.          *
*   Only do this once until the main loop recognizes it and clears  *
*   the "interrupt occurred" flag.  Clear the interrupt and leave.  *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
void ISR_lpt(void)
{
    if(!isrflag) {
        isrflag = 1;
        dirflag = (inportb(STATUS) & ROTENC_A) == 0;
    }
    outportb(picaddr,0x20); /* End of Interrupt (EOI) */
}
#endif   /* INTERRUPT */

/********************************************************************
* process_lpt()                                                     *
*                                                                   *
* Description:                                                      *
*   Updates seven-seg LED displays as required by mode, read inputs.*
*   Note: LED low-side drivers are open-collector pull-downs.  LED  *
*   is active when driver is LOW.                                   *
*                                                                   *
* Inputs  : int display, coded value to display on seven-seg        *
* Outputs : Return values as defined in global.h                    *
********************************************************************/
enum PARPORT_CONTROL process_lpt(int display)
{
    enum PARPORT_CONTROL retval;
    int swval;

    retval = CTLPANEL_NOACTION;

    /* Do mode-dependent setup for LED indicators. */
    /* Note: LED 1 (leftmost) control is NOT inverted.
     *    To turn Left LED (#1) ON, _clear_ the control bit
     *    To turn Left LED (#1) OFF, _set_ the control bit */
    if(((on_led1 == TRUE) || (lamp_set == DISPLAY_LAMP_TEST)) &&
            (lamp_set != DISPLAY_BLANK)) {
        shadow_control &= ~LLEDIND;
    }
    else {
        shadow_control |= LLEDIND;
    }
    if(((on_led2 == TRUE) || (lamp_set == DISPLAY_LAMP_TEST)) &&
            (lamp_set != DISPLAY_BLANK)) {
        shadow_control |= RLEDIND;
    }
    else {
        shadow_control &= ~RLEDIND;
    }

    /* Do mode-dependent setup for seven-segment display */
    shadow_port = LED_DRIVE;    /* Clear out PORT 7-seg drive bits */
    if((mode_7seg == SVNSEG_ON) || (lamp_set == DISPLAY_LAMP_TEST)) {
        /* Turn on all segments - regardless of number base */
        display = 0x88;
    }
    if((mode_7seg == SVNSEG_OFF) || (lamp_set == DISPLAY_BLANK)) {
        /* Turn off drive port bits */
        shadow_control &= ~SEVSEGS; 
    }
    else {
        /* Restart if coming out of "blank" mode */
        if((shadow_control & SEVSEGS) == 0) {
            shadow_control |= RSEVSEG;
        } 
        /* Flip the displays every time in here */
        shadow_control ^= SEVSEGS; 
        if((shadow_control & SEVSEGS) == RSEVSEG) {
            shadow_port |= cycle[(display & 0x000F)];
        }
        else {
            shadow_port |= cycle[(display & 0x00F0) >> 4];
        }
    }

    /* Done setting output bits.  Now update the hardware */
    outportb(PORT, shadow_port);
    outportb(CONTROL, shadow_control);

    /* Read status register for switch settings.  Note LOW is
     * switch DOWN (shorted to ground). */
    swval = inportb(STATUS) & SWITCHES;
    if((swval & LSWITCH) == 0) {
        switch(sw1_debounce) {
            case 0 :    /* Switch not detected before */
                sw1_debounce = 1;
                break;
            case 1:     /* Second sample of switch - CONFIRMED */
                state_sw1 = TRUE;
                retval = WIN_SW1_STATE;
                sw1_debounce = 2;
                break;
            default :   /* Only report switch closure ONCE */
                break;
        }
    }
    else {
        switch(sw1_debounce) {
            case 2 :    /* Switch not detected before */
                sw1_debounce = 1;
                break;
            case 1:     /* Second sample of switch - CONFIRMED */
                state_sw1 = FALSE;
                retval = WIN_SW1_STATE;
                sw1_debounce = 0;
                break;
            default :   /* Only report switch opening ONCE */
                break;
        }
    }
    if((swval & RSWITCH) == 0) {
        switch(sw2_debounce) {
            case 0 :    /* Switch not detected before */
                sw2_debounce = 1;
                break;
            case 1:     /* Second sample of switch - CONFIRMED */
                state_sw2 = TRUE;
                if(mode_sw2 != SW2_TERMINATE) {
                    retval = WIN_SW2_STATE;
                }
                else {
                    retval = CTLPANEL_TERMINATE;
                }
                sw2_debounce = 2;
                break;
            default :   /* Only report switch closure ONCE */
                break;
        }
    }
    else {
        switch(sw2_debounce) {
            case 2 :    /* Switch not detected before */
                sw2_debounce = 1;
                break;
            case 1:     /* Second sample of switch - CONFIRMED */
                state_sw2 = FALSE;
                retval = WIN_SW2_STATE;
                sw2_debounce = 0;
                break;
            default :   /* Only report switch opening ONCE */
                break;
        }
    }

#ifdef INTERRUPT
    /* Lowest priority: If a switch event is not detected, then check
     * if a Rotary Encoder interrupt has occurred. */
    if((retval == CTLPANEL_NOACTION) &&
            (mode_rotenc != ROTENC_DISABLE) &&
            (isrflag != 0)) {
        if(dirflag) {
            count_rotenc++;
        }
        else {
            count_rotenc--;
        }
        retval = WIN_ROTENC_COUNT;
    }
    isrflag = 0;
#else    /* Non-Interrupt */
    /* Lowest priority: Monitor rotary encoder channels for progression
     * through the states of two-bit Grey code.  If no change is detected,
     * remain in the current state.  Glitches will return to the previous
     * state until they settle out.  WHen a complete cycle is finished,
     * adjust the count variable. */
    if(retval == CTLPANEL_NOACTION) {
        swval = inportb(STATUS) & ROTENCS;
        switch(rotary_state) {
            case RSTATE_0 :
                switch(swval) {
                    case (ROTENC_A | ROTENC_B) :
                        /* Default state.  Stay here */
                        break;
                    case (ROTENC_A) :
                        /* Potential leftward rotation detected */
                        rotary_state = RSTATE_1L;
                        break;
                    case (ROTENC_B) :
                        /* Potential rightward rotation detected */
                        rotary_state = RSTATE_1R;
                        break;
                    case 0 :
                        /* Should never get here.  Ignore */
                        break;
                }
                break;
            case RSTATE_1L :
                switch(swval) {
                    case (ROTENC_A | ROTENC_B) :
                        /* Glitch.  Start over. */
                        rotary_state = RSTATE_0;
                        break;
                    case (ROTENC_A) :
                        /* Stable condition.  Stay here. */
                        break;
                    case (ROTENC_B) :
                        /* Should never get here.  Ignore */
                        break;
                    case 0 :
                        /* Leftward rotation continues */
                        rotary_state = RSTATE_2L;
                        break;
                }
                break;
            case RSTATE_2L :
                switch(swval) {
                    case (ROTENC_A | ROTENC_B) :
                        /* Should never get here.  Ignore */
                        break;
                    case (ROTENC_A) :
                        /* Glitch.  Back up one state. */
                        rotary_state = RSTATE_1L;
                        break;
                    case (ROTENC_B) :
                        /* Leftward rotation continues */
                        rotary_state = RSTATE_3L;
                        break;
                    case 0 :
                        /* Stable condition.  Stay here. */
                        break;
                }
                break;
            case RSTATE_3L :
                switch(swval) {
                    case (ROTENC_A | ROTENC_B) :
                        /* Cycle completed.  Move count and start over. */
                        count_rotenc++;
                        retval = WIN_ROTENC_COUNT;
                        rotary_state = RSTATE_0;
                        break;
                    case (ROTENC_A) :
                        /* Should never get here.  Ignore */
                        break;
                    case (ROTENC_B) :
                        /* Stable condition.  Stay here. */
                        break;
                    case 0 :
                        /* Glitch.  Back up one state. */
                        rotary_state = RSTATE_2L;
                        break;
                }
                break;
            case RSTATE_1R :
                switch(swval) {
                    case (ROTENC_A | ROTENC_B) :
                        /* Glitch.  Start over. */
                        rotary_state = RSTATE_0;
                        break;
                    case (ROTENC_A) :
                        /* Should never get here.  Ignore */
                        break;
                    case (ROTENC_B) :
                        /* Stable condition.  Stay here. */
                        break;
                    case 0 :
                        /* Rightward rotation continues */
                        rotary_state = RSTATE_2R;
                        break;
                }
                break;
            case RSTATE_2R :
                switch(swval) {
                    case (ROTENC_A | ROTENC_B) :
                        /* Should never get here.  Ignore */
                        break;
                    case (ROTENC_A) :
                        /* Rightward rotation continues */
                        rotary_state = RSTATE_3R;
                        break;
                    case (ROTENC_B) :
                        /* Glitch.  Back up one state. */
                        rotary_state = RSTATE_1R;
                        break;
                    case 0 :
                        /* Stable condition.  Stay here. */
                        break;
                }
                break;
            case RSTATE_3R :
                switch(swval) {
                    case (ROTENC_A | ROTENC_B) :
                        /* Cycle completed.  Move count and start over. */
                        count_rotenc--;
                        retval = WIN_ROTENC_COUNT;
                        rotary_state = RSTATE_0;
                        break;
                    case (ROTENC_A) :
                        /* Stable condition.  Stay here. */
                        break;
                    case (ROTENC_B) :
                        /* Should never get here.  Ignore */
                        break;
                    case 0 :
                        /* Glitch.  Back up one state. */
                        rotary_state = RSTATE_2R;
                        break;
                }
                break;
        }
    }
#endif   /* INTERRUPT */

    return retval;
}

