/********************************************************************
* parport.c                                                         *
*                                                                   *
* Description:                                                      *
*   Main module for PARPORT project.                                *
*                                                                   *
* Modification Record:                                              *
*                                                                   *
********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "global.h"
#include "ctlpanel.h"
#include "lpt_ctrl.h"

/************************ Global Constants *************************/
const int base_max[] = {             /* Maximum value for 2-digit display */
    100,                  /* DECIMAL, Base-10 */
    0x100,                /* HEXADECIMAL, Base-16 */
    0100                  /* OCTAL, Base-8 */
};

const int base_digit[] = {           /* Value of "10" in number base */
    10,                   /* DECIMAL, Base-10 */
    0x10,                 /* HEXADECIMAL, Base-16 */
    010                   /* OCTAL, Base-8 */
};

/************************ Global Variables *************************/
enum CTL_SW1_MODE mode_sw1;           /* Mode of Switch 1 */
enum CTL_SW2_MODE mode_sw2;           /* Mode of Switch 2 */
enum CTL_LED1_MODE mode_led1;         /* Mode of LED 1 */
enum CTL_LED2_MODE mode_led2;         /* Mode of LED 2 */
enum CTL_ROTENC_MODE mode_rotenc;     /* Mode of Rotary Encoder */
enum CTL_SVNSEG_MODE mode_7seg;       /* Mode of Seven-Seg Display */
int state_sw1;             /* State of Switch 1 (up/down) */
int toggle_sw1;            /* State toggled by Switch 1 (on/off) */
int enable_sw1;            /* TRUE to enable Switch 1 */
int state_sw2;             /* State of Switch 2 (up/down) */
int toggle_sw2;            /* State toggled by Switch 2 (on/off) */
int enable_sw2;            /* TRUE to enable Switch 2 */
int state_led1;            /* State of LED 1 */
int on_led1;               /* TRUE to turn LED 1 on */
int state_led2;            /* State of LED 2 */
int on_led2;               /* TRUE to turn LED 1 on */
int count_rotenc;          /* Count updated by Rotary Encoder */
int enable_rotenc;         /* TRUE to enable Rotary Encoder */
enum COUNT_BASE base_rotenc;   /* Counting base of Rotary Encoder */
int ulim_rotenc;           /* Upper limit for Rotary Encoder */
int llim_rotenc;           /* Lower limit for Rotary Encoder */
int display_7seg;          /* Value displayed by Seven-Seg Display */
enum COUNT_BASE base_7seg; /* Counting base of Seven-Seg Display */
enum DISPLAY_CONDITION lamp_set;   /* Display override flag */

/******************* Local Persistant Variables ********************/
static int display_7seg_coded; /* as display_7seg, but "coded" */
static clock_t old_time;       /* Previous "time" of 1-second update */
static clock_t now;            /* Current "time" */
static int tick_occurred;      /* Indicates a 54.9msec tick */
static int sec_occurred;       /* Indicates a second of time */
static int count_led1;         /* Timer for LED1 blink/ramp */
static int count_led2;         /* Timer for LED2 blink/ramp */
static int thresh_led1;        /* Threshold for LED1 ramp */
static int thresh_led2;        /* Threshold for LED2 ramp */
static int direct_led1;        /* Direction for LED1 ramp */
static int direct_led2;        /* Direction for LED2 ramp */

/************************** Local Symbols **************************/
#define MAX_BLINK    7          /* Blink rate for LED in Blink mode */
#define RAMP_CYCLE  10          /* Cycle time for ramp in Ramp mode */

/************************ Local Prototypes *************************/
void init_vars(void);
void handle_gui_event(enum PARPORT_CONTROL ctl);
void handle_hw_event(enum PARPORT_CONTROL ctl);
void update_sevenseg(void);
void update_led(void);
int int_to_coded(int x, enum COUNT_BASE b);

/********************************************************************
* main()                                                            *
*                                                                   *
* Description:                                                      *
*   Top-level execution function.  Initialize all CSUs, run main    *
*   loop, check for terminate condition, and shut down.             *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
int main(void)
{
    enum PARPORT_CONTROL ctl;
    clock_t temp_time;

    /* Call initialization routines */
    init_vars();
    init_parport();
    init_ctlpanel();

    /* MAIN LOOP */
    while(1) {
        /* Get current time.  If a tick or a second has passed,
         * set the indicators. */
        temp_time = clock();
        if(now != temp_time) {
            now = temp_time;
            tick_occurred = TRUE;
        }
        if((now - old_time) >= CLOCKS_PER_SEC) {
            old_time = now;
            sec_occurred = TRUE;
        }

        /* Update variables for seven-segment display 
         * and LED indicators */
        update_sevenseg();
        update_led();

        /* Process control panel events */
        ctl = process_ctlpanel();
        if(ctl == CTLPANEL_TERMINATE)
            break;
        handle_gui_event(ctl);

        /* Process hardware module events */
        ctl = process_lpt(display_7seg_coded);
        if(ctl == CTLPANEL_TERMINATE)
            break;
        handle_hw_event(ctl);

        /* Reset time indicators */
        tick_occurred = 0;
        sec_occurred = 0;
    }

    /* Call shutdown routines */
    shutdown_ctlpanel();
    shutdown_parport();

    putchar('\n');    /* Scroll the screen up one line */
    return 0;
}

/********************************************************************
* init_vars()                                                       *
*                                                                   *
* Description:                                                      *
*   Initialize global variables.                                    *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
void init_vars(void)
{
    /* Globals */
    mode_sw1 = SW1_TOGGLE_LED1;
    mode_sw2 = SW2_TOGGLE_LED2;
    mode_led1 = LED1_NORMAL;
    mode_led2 = LED2_NORMAL;
    mode_rotenc = ROTENC_NORMAL;
    mode_7seg = SVNSEG_TIMER;
    state_sw1 = 0;
    toggle_sw1 = 0;
    enable_sw1 = 0;
    state_sw2 = 0;
    toggle_sw2 = 0;
    enable_sw2 = 0;
    state_led1 = 0;
    on_led1 = 0;
    state_led2 = 0;
    on_led2 = 0;
    count_rotenc = 0;
    enable_rotenc = 0;
    base_rotenc = DECIMAL;
    ulim_rotenc = 0;
    llim_rotenc = 0;
    display_7seg = 0;
    base_7seg = DECIMAL;
    lamp_set = DISPLAY_NORMAL;

    /* Local statics */
    display_7seg_coded = 0;
    old_time = clock();
    now = old_time;
    tick_occurred = 0;
    sec_occurred = 0;
    count_led1 = 0;
    count_led2 = 0;
    thresh_led1 = 0;
    thresh_led2 = 0;
    direct_led1 = 1;
    direct_led2 = 1;
}

/********************************************************************
* handle_gui_event()                                                *
*                                                                   *
* Description:                                                      *
*   When certain control panel events occur, some states may need   *
*   to be updated.                                                  *
*                                                                   *
* Inputs  : enum PARPORT_CONTROL ctl, ID of control associated with *
*             event                                                 *
* Outputs : None                                                    *
********************************************************************/
void handle_gui_event(enum PARPORT_CONTROL ctl)
{
    switch(ctl) {
        case WIN_ROTENC_RESET :
            if(mode_rotenc == ROTENC_LIMIT) {
                count_rotenc = llim_rotenc;
            }
            else {
                count_rotenc = 0;
            }
            update_ctlpanel(WIN_ROTENC_COUNT);
            break;
        case WIN_ROTENC_ULIMIT :
            /* If in Limit mode, make sure count is within limits */
            if((mode_rotenc == ROTENC_LIMIT) && (count_rotenc > ulim_rotenc)) {
                count_rotenc = ulim_rotenc;
                update_ctlpanel(WIN_ROTENC_COUNT);
            }
            break;
        case WIN_ROTENC_LLIMIT :
            /* If in Limit mode, make sure count is within limits */
            if((mode_rotenc == ROTENC_LIMIT) && (count_rotenc < llim_rotenc)) {
                count_rotenc = llim_rotenc;
                update_ctlpanel(WIN_ROTENC_COUNT);
            }
            break;
        case WIN_ROTENC_BASE :
            /* Make sure all numeric parameters are within range of new
             * numeric base, and then (recursively) fix count as required */
            count_rotenc %= base_max[base_rotenc];
            ulim_rotenc %= base_max[base_rotenc];
            llim_rotenc %= base_max[base_rotenc];
            update_ctlpanel(WIN_ROTENC_COUNT);
            update_ctlpanel(WIN_ROTENC_ULIMIT);
            update_ctlpanel(WIN_ROTENC_LLIMIT);
            handle_gui_event(WIN_ROTENC_ULIMIT);
            handle_gui_event(WIN_ROTENC_LLIMIT);
            break;
        case WIN_ROTENC_MODE :
            /* If mode changes to "Limited", (recursively) make sure
             * the count is within the limits */
            if(mode_rotenc == ROTENC_LIMIT) {
                handle_gui_event(WIN_ROTENC_ULIMIT);
                handle_gui_event(WIN_ROTENC_LLIMIT);
            }
            break;
        case WIN_7SEG_BASE :
            /* Make sure all numeric parameters are within range of new
             * numeric base. */
            display_7seg %= base_max[base_7seg];
            update_ctlpanel(WIN_7SEG_VALUE);
            break;
        default :        /* Event that requires no special handling */
            break;
    }
}

/********************************************************************
* handle_hw_event()                                                 *
*                                                                   *
* Description:                                                      *
*   Process events detected from hardware module.                   *
*                                                                   *
* Inputs  : enum PARPORT_CONTROL ctl, ID of control associated with *
*             event                                                 *
* Outputs : None                                                    *
********************************************************************/
void handle_hw_event(enum PARPORT_CONTROL ctl)
{
    switch(ctl) {
        case WIN_SW1_STATE :
            /* Only process switch DOWN events */
            if(state_sw1) {
                switch(mode_sw1) {
                    case SW1_NORMAL :       /* Only update control panel */
                        break;
                    case SW1_TOGGLE_LED1 :  /* Change LED1 state */
                        state_led1 = (state_led1) ? FALSE : TRUE;
                        break;
                    case SW1_TOGGLE_7SEG :  /* Change 7-seg mode */
                        mode_7seg++;
                        if(mode_7seg == MAX_SVNSEG_MODE) {
                            mode_7seg = 0;
                        }
                        update_ctlpanel(WIN_7SEG_MODE);
                        break;
                    case SW1_DISABLE :      /* Do nothing */
                    default :
                        return;
                        break;
                }
                toggle_sw1 = (toggle_sw1) ? FALSE : TRUE;
                update_ctlpanel(WIN_SW1_STATE);
            }
            break;
        case WIN_SW2_STATE :
            /* Only process switch DOWN events */
            if(state_sw2) {
                switch(mode_sw2) {
                    case SW2_NORMAL :       /* Only update control panel */
                        break;
                    case SW2_TOGGLE_LED2 :  /* Change LED1 state */
                        state_led2 = (state_led2) ? FALSE : TRUE;
                        break;
                    case SW2_TOGGLE_DISPLAY :   /* Blank all displays */
                        lamp_set++;
                        if(lamp_set == MAX_DISPLAY_CONDITION) {
                            lamp_set = 0;
                        }
                        break;
                    case SW2_TERMINATE :    /* Halt program */
                        /* Already handled by lpt_ctrl.c */
                        break;
                    case SW2_DISABLE :      /* Do nothing */
                    default :
                        return;
                        break;
                }
                toggle_sw2 = (toggle_sw2) ? FALSE : TRUE;
                update_ctlpanel(WIN_SW2_STATE);
            }
            break;
        case WIN_ROTENC_COUNT :
            switch(mode_rotenc) {
                case ROTENC_NORMAL :
                    /* Spin around endpoint for the base.
                     * Always update control panel */
                    if(count_rotenc >= base_max[base_rotenc]) {
                        count_rotenc = 0;
                    }
                    else if(count_rotenc < 0) {
                        count_rotenc = base_max[base_rotenc] - 1;
                    }
                    update_ctlpanel(WIN_ROTENC_COUNT);
                    break;
                case ROTENC_LIMIT :
                    /* Limit count to min/max.  Only need to
                     * update control panel if not at limit already */
                    if(count_rotenc > ulim_rotenc) {
                        count_rotenc = ulim_rotenc;
                    }
                    else if(count_rotenc < llim_rotenc) {
                        count_rotenc = llim_rotenc;
                    }
                    else {
                        update_ctlpanel(WIN_ROTENC_COUNT);
                    }
                    break;
                default :
                    /* Already handled by lpt_ctrl.c */
                    break;
            }
            break;
        default :
            break;
    }
}

/********************************************************************
* update_sevenseg()                                                 *
*                                                                   *
* Description:                                                      *
*   Update variables associated with seven-segment display as       *
*   conditions require.                                             *
*                                                                   *
* Inputs  : None.                                                   *
* Outputs : None.                                                   *
********************************************************************/
void update_sevenseg(void)
{
    int update_reqd;

    update_reqd = FALSE;

    switch(mode_7seg) {
        case SVNSEG_TIMER :
            /* Update timer display once a second */
            if(sec_occurred) {
                display_7seg++;

                /* and handle base-unique rollover */
                if(display_7seg >= base_max[base_7seg]) {
                    display_7seg = 0;
                }
                update_reqd = TRUE;
            }
            break;
        case SVNSEG_ROTENC :
            /* Keep 7-seg display in lock-step with rotary encoder */
            if(display_7seg != count_rotenc) {
                if(count_rotenc >= base_max[base_7seg]) {
                    display_7seg = count_rotenc % base_max[base_7seg];
                }
                else {
                    display_7seg = count_rotenc;
                }
                update_reqd = TRUE;
            }
            break;
        default :         /* Don't care about other modes */
            break;
    }

    /* If something has occurred to change the display value,
     * encode it for 7-segment display, and mirror it on the
     * control panel. */
    if(update_reqd) {
        display_7seg_coded = int_to_coded(display_7seg, base_7seg);
        update_ctlpanel(WIN_7SEG_VALUE);
    }
}

/********************************************************************
* update_led()                                                      *
*                                                                   *
* Description:                                                      *
*   Turn on/off LED indicators per mode.                            *
*                                                                   *
* Inputs  : None.                                                   *
* Outputs : None.                                                   *
********************************************************************/
void update_led(void)
{
    switch(mode_led1) {
        case LED1_NORMAL :
            on_led1 = state_led1;
            break;
        case LED1_BLINK :
            if(state_led1 == TRUE) {
                /* Blink: count 54.9ms DOS "ticks" until half-cycle
                 * threshold reached, then toggle LED state and
                 * reset count. */
                if(tick_occurred) {
                    count_led1++;
                    if(count_led1 >= MAX_BLINK) {
                        on_led1 = (on_led1) ? FALSE : TRUE;
                        count_led1 = 0;
                    }
                }
            }
            else {
                on_led1 = FALSE;
            }
            break;
        case LED1_RAMP :
            if(state_led1 == TRUE) {
                /* Ramp ("Soft Blink"): Use count_ledx as a sawtooth
                 * generator - count from zero to max and then
                 * reset to zero.  Generate a Pulse-Width Modulation
                 * output for the LED control by comparing the
                 * sawtooth with a threshold; if the sawtooth is
                 * less than threshold, turn the LED on, otherwise,
                 * it's off (thus, when threshold starts at zero,
                 * the LED is off all the time).  Then, modulate
                 * the pulse width by using the threshold as a
                 * triangle wave - every 54.9ms DOS "tick", count
                 * from zero to max, and when max is reached, count
                 * back down to zero.  This scheme works as long as
                 * the PWM cycle (counts every pass through the main
                 * loop) is much shorter than the threshold adjustment, 
                 * and since even slow computers will run hundreds or
                 * thousands of lines of code in 54.9ms, this should
                 * be a good assumption.  The max threshold is thus
                 * the half-cycle of the intensity or "soft blink"
                 * profile for the LED. */
                if(tick_occurred) {
                    thresh_led1 += direct_led1;
                    if(thresh_led1 == RAMP_CYCLE) {
                        direct_led1 = -1;
                    }
                    else if(thresh_led1 == 0) {
                        direct_led1 = 1;
                    }
                }
                if(count_led1 < thresh_led1) {
                    on_led1 = TRUE;
                }
                else {
                    on_led1 = FALSE;
                }
                count_led1++;
                if(count_led1 >= RAMP_CYCLE) {
                    count_led1 = 0;
                }
            }
            else {
                on_led1 = FALSE;
            }
            break;
        case LED1_ON :
            on_led1 = TRUE;
            break;
        case LED1_OFF :
            on_led1 = FALSE;
            break;
        default :     /* Never gets here.  Make the compiler happy */
            break;
    }

    /* LED2 same logic as LED1.  Might be more efficient if LED controls
     * were arrays (since modes of both are identical).  Oh, well... */
    switch(mode_led2) {
        case LED2_NORMAL :
            on_led2 = state_led2;
            break;
        case LED2_BLINK :
            if(state_led2 == TRUE) {
                if(tick_occurred) {
                    count_led2++;
                    if(count_led2 >= MAX_BLINK) {
                        on_led2 = (on_led2) ? FALSE : TRUE;
                        count_led2 = 0;
                    }
                }
            }
            else {
                on_led2 = FALSE;
            }
            break;
        case LED2_RAMP :
            if(state_led2 == TRUE) {
                if(tick_occurred) {
                    thresh_led2 += direct_led2;
                    if(thresh_led2 == RAMP_CYCLE) {
                        direct_led2 = -1;
                    }
                    else if(thresh_led2 == 0) {
                        direct_led2 = 1;
                    }
                }
                if(count_led2 < thresh_led2) {
                    on_led2 = TRUE;
                }
                else {
                    on_led2 = FALSE;
                }
                count_led2++;
                if(count_led2 >= RAMP_CYCLE) {
                    count_led2 = 0;
                }
            }
            else {
                on_led2 = FALSE;
            }
            break;
        case LED2_ON :
            on_led2 = TRUE;
            break;
        case LED2_OFF :
            on_led2 = FALSE;
            break;
        default :     /* Never gets here.  Make the compiler happy */
            break;
    }
}

/********************************************************************
* int_to_coded()                                                    *
*                                                                   *
* Description:                                                      *
*   Convert an integer into a "Binary Coded Decimal" number, except *
*   not just "decimal", but whatever base is supplied.              *
*   Note: input is "int", but only works for unsigned values.       *
*                                                                   *
* Inputs  : int x, number to be converted                           *
*           enum COUNT_BASE b, base to convert number into          *
* Outputs : int, coded number                                       *
********************************************************************/
int int_to_coded(int x, enum COUNT_BASE b)
{
    int retval;

    /* Don't need to do anything for HEXADECIMAL, since hex is
     * the native representation base for binary computers. */
    if(b == HEXADECIMAL) {
        return x;
    }

    retval = x % base_digit[b];     /* LS digit. */
    x = x / base_digit[b];          /* Move MS digit to LS.  Also drops
                                     * LS digit since no rounding occurs */
    x = x % base_digit[b];          /* Constrain MS digit to a single digit
                                     * in the desired base.  This means a
                                     * third digit will be lost, but that
                                     * should have been taken care of outside
                                     * of this function! */
    retval |= x << 4;               /* Move digit to MS nybble */
    return retval;
}

