/********************************************************************
* rotenc.c                                                          *
*                                                                   *
* Description:                                                      *
*   With a rotary encoder connected with Channel A -> PAPER_OUT     *
*   (Status Register bit 5) and Channel B -> ACK (Status Register   *
*   bit 6) and the common pin to ground, monitor the channels as    *
*   the encoder is turned, and report.  In addition, enable the ACK *
*   interrupt (Channel B) and report the counts of Channel A being  *
*   high (for one direction of turning) and low (for the other      *
*   direction).                                                     *
*                                                                   *
* Modification Record:                                              *
*   27Dec09 DL Ormand     Initial Release                           *
********************************************************************/
/* Based on:                                                           */
/*  Parallel Port Interrupt Polarity Tester                            */
/*  2nd February 1998                                                  */
/*  Copyright 1997 Craig Peacock                                       */
/*  WWW      - http://www.senet.com.au/~cpeacock                       */
/*  Email    - cpeacock@senet.com.au                                   */
#include <stdio.h>
#include <stdlib.h>
#include <pc.h>
#include <dpmi.h>
#include <go32.h>
#include <dos.h>

#define PORTADDRESS 0x378   /* Enter Your Port Address Here */
#define IRQ 7               /* IRQ Here */
#define DATA PORTADDRESS+0
#define STATUS PORTADDRESS+1
#define CONTROL PORTADDRESS+2
#define PIC1 0x20
#define PIC2 0xA0

/* Macros by Shawn Hargreaves from the Allegro library for locking
 * functions and variables. */
#define LOCK_VARIABLE(x)    _go32_dpmi_lock_data((void *)&x,(long)sizeof(x));
#define LOCK_FUNCTION(x)    _go32_dpmi_lock_code(x,(long)sizeof(x));

int interflag;  /* Interrupt Flag */
int count_one;  /* Channel A High counter */
int count_zero; /* Channel A Low counter */
int picaddr;     /* Programmable Interrupt Controller (PIC) Base Address */

/* Interrupt handler context storage */
_go32_dpmi_seginfo OldISR, NewISR;

/********************************************************************
* parisr()                                                          *
*                                                                   *
* Description:                                                      *
*   Parallel Port Interrupt Service Routine                         *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : None                                                    *
********************************************************************/
void parisr(void) /* Interrupt Service Routine (ISR) */
{
   interflag = 1;
   /* Look at the condition of PAPER_OUT (bit 5), count a 1 if
    * high, count a 0 if low */
   if(inportb(STATUS) & 0x20) {
       count_one++;
   }
   else {
       count_zero++;
   }
   outportb(picaddr,0x20); /* End of Interrupt (EOI) */
}

/********************************************************************
* main()                                                            *
*                                                                   *
* Description:                                                      *
*   Top-level execution.  Set up interrupt, continuously monitor    *
*   PAPER_OUT and ACK and report when a change occurs, along with   *
*   interrupt activity, until a keypress is detected.  Then restore *
*   original interrupt context and exit.                            *
*                                                                   *
* Inputs  : None                                                    *
* Outputs : Always returns 0                                        *
********************************************************************/
int main(void)
{
  int old, new;
  int ch_a, ch_b;
  int nl_req;
  int intno;      /* Interrupt Number */
  int picmask;    /* PIC's Mask */
  /* Calculate Interrupt Number, PIC Addr & Mask. */
  if(IRQ >= 2 && IRQ <= 7) {
    intno = IRQ + 0x08;
    picaddr = PIC1;
    picmask = 1;
    picmask = picmask << IRQ;
  }
  if(IRQ >= 8 && IRQ <= 15) {
    intno = IRQ + 0x68;
    picaddr = PIC2;
    picmask = 1;
    picmask = picmask << (IRQ-8);
  }
  if(IRQ < 2 || IRQ > 15) {
    printf("IRQ Out of Range\n");
    exit(0);
  }

  /* lock the functions and variables */
  LOCK_FUNCTION(parisr);
  LOCK_VARIABLE(interflag);
  LOCK_VARIABLE(count_one);
  LOCK_VARIABLE(count_zero);
  LOCK_VARIABLE(picaddr);

  /* load the address of the old timer ISR into the OldISR structure */
  _go32_dpmi_get_protected_mode_interrupt_vector(intno, &OldISR);
  /* point NewISR to the proper selector:offset for handler
   * function */
  NewISR.pm_offset = (int)parisr;
  NewISR.pm_selector = _go32_my_cs();
  _go32_dpmi_allocate_iret_wrapper(&NewISR);
  _go32_dpmi_set_protected_mode_interrupt_vector(intno,&NewISR);
  outportb(picaddr+1,inportb(picaddr+1) & (0xFF - picmask)); /* Un-Mask Pic */
  outportb(CONTROL, inportb(CONTROL) | 0x10); /* Enable Parallel Port IRQ's */

  /* Start execution with a greeting, and set up variables */
  system("CLS");
  printf("Rotary Encoder Channel Tester\n");
  printf("IRQ %d : INTNO %02X : PIC Addr 0x%X : Mask 0x%02X\n",
      IRQ, intno, picaddr, picmask);
  interflag = 0; /* Reset Interrupt Flag */
  count_one = 0;
  count_zero = 0;
  new = inportb(STATUS) & 0x60;
  nl_req = 0;

  /* Main Loop: Continue until keypress */
  while(!kbhit()) {
      /* Get condition of PAPER_OUT and ACK lines */
      new = inportb(STATUS) & 0x60;
      ch_a = new & 0x20;
      ch_b = new & 0x40;

      /* If a change has occurred, report the line conditions */
      if(new != old) {
          printf("%d\t%d",
                  (ch_a) ? 1 : 0,
                  (ch_b) ? 1 : 0);
          old = new;
          nl_req = 1;
      }

      /* Additionally, if an interrupt has occurred, report the counts */
      if(interflag) {
          printf("\t%d\t%d", count_one, count_zero);
          interflag = 0;
          nl_req = 1;
      }

      /* If any report has been made, end the screen line */
      if(nl_req) {
          printf("\n");
          nl_req = 0;
      }
  }

  /* Disable Parallel Port IRQ's */
  outportb(CONTROL, inportb(CONTROL) & 0xEF);
  outportb(picaddr+1,inportb(picaddr+1) | picmask); /* Mask Pic */
  /* load the old timer ISR back without the new ISR chained on */
  _go32_dpmi_set_protected_mode_interrupt_vector(intno,&OldISR);

  return 0;
}

